from importlib import metadata

CATWALK_CLIENT_VERSION = metadata.version("catwalk_client")

CATWALK_AUTH_HEADER = "Catwalk-Authorization"
CATWALK_USER_AGENT_HEADER_VALUE = f"Catwalk-Client/{CATWALK_CLIENT_VERSION}"



from datetime import datetime, timedelta
from json import loads
from logging import getLogger
from time import sleep

from catwalk_common import OpenCase

from catwalk_client.common import CatwalkHTTPClient

logger = getLogger("catwalk_client")


class CaseExporter:
    def __init__(
        self,
        http_client: CatwalkHTTPClient,
        from_datetime: datetime,
        to_datetime: datetime,
        submitter_name: str = None,
        submitter_version: str = None,
        max_retries: int = 5,
        batch: int = 1,
    ):
        self.http_client = http_client

        self.from_datetime = from_datetime
        self.to_datetime = to_datetime

        self.filters = self._build_filters(
            submitter_name, submitter_version, from_datetime, to_datetime
        )
        self.max_retries = max_retries
        self.batch = batch

    def export(self):
        for date in self._hour_range():
            current_retry = 0
            path = f"/api/cases/export/{date.year}/{date.month}/{date.day}/{date.hour}"

            while path:
                response, success = self.http_client.get(path, self.filters)

                if not success:
                    self._handle_retry(current_retry, response)
                    current_retry += 1
                    continue

                path = None
                data = loads(response)
                for item in data["items"]:
                    yield OpenCase.parse_obj(item)

                if data["next_part"] and data["items"]:
                    path = data["next_part"]

    def _handle_retry(self, current_retry: int, response):
        if current_retry < self.max_retries:
            multiplier = (current_retry * 2) or 1
            logger.error(
                f"\n[catwalk-export] Retry ({current_retry + 1}/{self.max_retries}) "
                f"because got error {response} "
            )
            sleep(0.5 * multiplier)
        else:
            raise Exception("Max retries exceeded")

    @staticmethod
    def _build_filters(
        submitter_name: str,
        submitter_version: str,
        from_datetime: datetime,
        to_datetime: datetime,
    ):
        filters = {
            "submitter_name": submitter_name,
            "submitter_version": submitter_version,
            "from_timestamp": from_datetime.isoformat(),
            "to_timestamp": to_datetime.isoformat(),
        }
        return {k: v for k, v in filters.items() if v is not None}

    def _get_url(self, path: str):
        return self.catwalk_url.rstrip("/") + path

    def _hour_range(self):
        start, end = self.from_datetime, self.to_datetime

        while start < end:
            yield start
            start += timedelta(hours=1)



from __future__ import annotations

from typing import Any, Union


class CaseBuilder:
    def __init__(self, client):
        self._client = client
        self._case_id = ""
        self._query: list[dict] = []
        self._context: list[dict] = []
        self._response: list[dict] = []
        self._metadata: dict = {}

    def set_case_id(self, case_id: str):
        self._case_id = case_id
        return self

    def add_query(self, name: str, value: Any, type: Union[str, dict]):
        self._query.append({"name": name, "value": value, "type": type})
        return self

    def add_context(self, name: str, value: Any, type: Union[str, dict]):
        self._context.append({"name": name, "value": value, "type": type})
        return self

    def add_response(
        self,
        name: str,
        value: Any,
        type: Union[str, dict],
        evaluation: list[dict] = None,
    ):
        self._response.append(
            {"name": name, "value": value, "type": type, "evaluation": evaluation or []}
        )
        return self

    def add_evaluation(self, question: str, name: str, **kwargs):
        self._response[-1]["evaluation"].append(
            {"question": question, "name": name, **kwargs}
        )
        return self

    def set_metadata(self, **kwargs):
        self._metadata = kwargs
        return self

    def send(self):
        self._client.send(
            self._case_id,
            {
                "query": self._query,
                "context": self._context,
                "response": self._response,
                "metadata": self._metadata,
            },
        )



import logging

from datetime import datetime
from json import loads

from catwalk_common import CommonCaseFormat
from pydantic import error_wrappers

from catwalk_client.tools import CaseBuilder, CaseExporter
from catwalk_client.common import CatwalkHTTPClient


logger = logging.getLogger("catwalk_client")
logging.basicConfig(level=logging.INFO)


class CatwalkClient:
    def __init__(
        self,
        submitter_name: str = None,
        submitter_version: str = None,
        catwalk_url: str = None,
        auth_token: str = None,
        insecure: bool = True,
    ):
        self.http_client = CatwalkHTTPClient(catwalk_url, auth_token, insecure)
        self.submitter_name = submitter_name
        self.submitter_version = submitter_version

    def __setattr__(self, __name, __value):
        if __name != "http_client" and __name in self.http_client.__dict__.keys():
            logger.warn(
                f" [DEPRECATED] Usage of 'CatwalkClient.{__name}=<value>' is DEPRECATED. Please use 'set_{__name}' method!"
            )
            self.http_client.__dict__[__name] = __value

        self.__dict__[__name] = __value

    def set_auth_token(self, auth_token: str):
        self.http_client.auth_token = auth_token

    def set_catwalk_url(self, catwalk_url: str):
        self.http_client.catwalk_url = catwalk_url

    def set_insecure(self, insecure: bool):
        self.http_client.insecure = insecure

    def new_case(self) -> CaseBuilder:
        return CaseBuilder(client=self)

    def send(self, case: dict):
        try:
            case = CommonCaseFormat(
                submitter={
                    "name": self.submitter_name,
                    "version": self.submitter_version,
                },
                **case,
            )

            response, success = self.http_client.post("/api/cases/collect", case.json())

            if success:
                case_id = loads(response)["id"]
                logger.info(f" Collected catwalk case: {case_id}")
            else:
                logger.error(response)
        except error_wrappers.ValidationError as ex:
            logger.error(f" ValidationError: \n{ex}")
        except Exception as ex:
            logger.error(f" {type(ex).__name__}: \n{str(ex)}")

    def export_cases(
        self,
        from_datetime: datetime,
        to_datetime: datetime,
        submitter_name: str = None,
        submitter_version: str = None,
        max_retries: int = 5,
    ):
        exporter = CaseExporter(
            http_client=self.http_client,
            from_datetime=from_datetime,
            to_datetime=to_datetime,
            submitter_name=submitter_name or self.submitter_name,
            submitter_version=submitter_version or self.submitter_version,
            max_retries=max_retries,
        )
        yield from exporter.export()



from os import environ

from urllib.request import Request

from ._http_client import HTTPClient
from .constants import (
    CATWALK_AUTH_HEADER,
    CATWALK_USER_AGENT_HEADER_VALUE,
)


class CatwalkHTTPClient(HTTPClient):
    def __init__(self, catwalk_url: str, auth_token: str, insecure: bool = True):
        super().__init__(
            catwalk_url or environ.get("CATWALK_URL"),
            auth_token or environ.get("CATWALK_AUTH_TOKEN"),
            insecure,
        )

    def _apply_headers(self, request: Request):
        self._add_auth_token_header(request, self.auth_token)
        self._add_user_agent_header(request, CATWALK_USER_AGENT_HEADER_VALUE)

    def _add_auth_token_header(self, request: Request, header_value: str = ""):
        request.add_header(CATWALK_AUTH_HEADER, f"Bearer {header_value}")



from __future__ import annotations

from json import loads
from ssl import SSLContext, CERT_NONE, create_default_context
from typing import Any, Optional

from urllib.error import URLError, HTTPError
from urllib.parse import urlencode
from urllib.request import urlopen, Request


class HTTPClient:
    def __init__(self, url: str, auth_token: str, insecure: bool = True):
        self.url = url
        self.auth_token = auth_token
        self.insecure = insecure

    def get_url(self, path: str):
        return self.url.rstrip("/") + path

    def get(
        self, url_postfix: str, query_params: Optional[dict[str, Any]] = None
    ) -> tuple[str, bool]:
        url = self.get_url(url_postfix) + self._encode_query_params(query_params)
        req = Request(url=url)
        return self._make_request(req)

    def post(
        self,
        url_postfix: str,
        payload: str,
        query_params: Optional[dict[str, Any]] = None,
    ) -> tuple[str, bool]:
        url = self.get_url(url_postfix) + self._encode_query_params(query_params)
        req = Request(url=url, data=payload.encode())
        req.add_header("Content-Type", "application/json")
        return self._make_request(req)

    def _make_request(self, request: Request):
        self._apply_headers(request)
        try:
            resp = urlopen(request, timeout=5, context=self._get_ssl_context())
        except HTTPError as e:
            e_body = self._parse_http_error_json(e)
            return f"Error: {e}\n[URL] {e.geturl()}\n[DETAILS] {e_body}", False
        except URLError as e:
            return f"Connection error: {e.reason}", False

        return resp.read().decode(), True

    def _parse_http_error_json(self, error: HTTPError):
        if error.headers.get_content_type() == "application/json":
            return loads(error.read().decode())
        return None

    def _get_ssl_context(self) -> SSLContext:
        ctx = create_default_context()
        if self.insecure:
            ctx.check_hostname = False
            ctx.verify_mode = CERT_NONE
        return ctx

    def _encode_query_params(self, query_params: Optional[dict[str, Any]]) -> str:
        if query_params is None or len(query_params) == 0:
            return ""
        encoded_params = urlencode(query_params)
        return "?" + encoded_params

    def _apply_headers(self, request: Request):
        self._add_user_agent_header(request)
        self._add_auth_token_header(request, self.auth_token)

    def _add_auth_token_header(self, request: Request, header_value: str = ""):
        request.add_header("Authorization", f"Bearer {header_value}")

    def _add_user_agent_header(
        self, request: Request, header_value: str = "Request/1.0"
    ):
        if not request.has_header("User-Agent"):
            request.add_header("User-Agent", header_value)
